from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.auth.models import Group
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.shortcuts import get_object_or_404, redirect
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.views.generic import CreateView, DeleteView, DetailView, UpdateView
from django_filters.views import FilterView
from django_tables2 import SingleTableMixin

from hypha.apply.activity.messaging import MESSAGES, messenger
from hypha.apply.activity.models import APPLICANT, COMMENT, Activity
from hypha.apply.todo.options import (
    INVOICE_REQUIRED_CHANGES,
    INVOICE_WAITING_APPROVAL,
    PROJECT_WAITING_INVOICE,
)
from hypha.apply.todo.views import (
    add_task_to_user_group,
    remove_tasks_for_user,
    remove_tasks_for_user_group,
    remove_tasks_of_related_obj,
)
from hypha.apply.users.decorators import staff_or_finance_required
from hypha.apply.users.groups import STAFF_GROUP_NAME
from hypha.apply.utils.storage import PrivateMediaView
from hypha.apply.utils.views import DelegateableView, DelegatedViewMixin, ViewDispatcher

from ..filters import InvoiceListFilter
from ..forms import ChangeInvoiceStatusForm, CreateInvoiceForm, EditInvoiceForm
from ..models.payment import (
    APPROVED_BY_FINANCE,
    APPROVED_BY_STAFF,
    CHANGES_REQUESTED_BY_FINANCE,
    CHANGES_REQUESTED_BY_STAFF,
    INVOICE_TRANISTION_TO_RESUBMITTED,
    Invoice,
)
from ..models.project import PROJECT_ACTION_MESSAGE_TAG, Project
from ..services import handle_tasks_on_invoice_update
from ..tables import InvoiceListTable


@method_decorator(login_required, name="dispatch")
class InvoiceAccessMixin(UserPassesTestMixin):
    model = Invoice

    def get_object(self):
        project = get_object_or_404(Project, pk=self.kwargs["pk"])
        return get_object_or_404(project.invoices.all(), pk=self.kwargs["invoice_pk"])

    def test_func(self):
        if self.request.user.is_apply_staff:
            return True

        if self.request.user.is_finance:
            return True

        if self.request.user == self.get_object().project.user:
            return True

        return False


@method_decorator(staff_or_finance_required, name="dispatch")
class ChangeInvoiceStatusView(DelegatedViewMixin, InvoiceAccessMixin, UpdateView):
    form_class = ChangeInvoiceStatusForm
    context_name = "change_invoice_status"
    model = Invoice

    def form_valid(self, form):
        invoice = get_object_or_404(
            Invoice, pk=self.kwargs["invoice_pk"]
        )  # to get the old status
        old_status = invoice.status
        response = super().form_valid(form)
        if form.cleaned_data["comment"]:
            invoice_status_change = _(
                "<p>Invoice status updated to: {status}.</p>"
            ).format(status=self.object.get_status_display())
            comment = f"<p>{self.object.comment}.</p>"

            message = invoice_status_change + comment

            Activity.objects.create(
                user=self.request.user,
                type=COMMENT,
                source=self.object.project,
                timestamp=timezone.now(),
                message=message,
                visibility=APPLICANT,
                related_object=self.object,
            )

        if (
            self.request.user.is_apply_staff and self.object.status == APPROVED_BY_STAFF
        ) or (
            settings.INVOICE_EXTENDED_WORKFLOW
            and self.request.user.is_finance_level_1
            and self.object.status == APPROVED_BY_FINANCE
        ):
            messenger(
                MESSAGES.APPROVE_INVOICE,
                request=self.request,
                user=self.request.user,
                source=self.object.project,
                related=self.object,
            )

        messenger(
            MESSAGES.UPDATE_INVOICE_STATUS,
            request=self.request,
            user=self.request.user,
            source=self.object.project,
            related=self.object,
        )

        handle_tasks_on_invoice_update(old_status=old_status, invoice=self.object)

        return response


class DeleteInvoiceView(DeleteView):
    model = Invoice

    def get_object(self):
        project = get_object_or_404(Project, pk=self.kwargs["pk"])
        return get_object_or_404(project.invoices.all(), pk=self.kwargs["invoice_pk"])

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        if not self.object.can_user_delete(request.user):
            raise PermissionDenied

        return super().dispatch(request, *args, **kwargs)

    @transaction.atomic()
    def form_valid(self, form):
        # remove all tasks related to this invoice irrespective of code and users/user_group
        remove_tasks_of_related_obj(related_obj=self.object)

        response = super().form_valid(form)

        messenger(
            MESSAGES.DELETE_INVOICE,
            request=self.request,
            user=self.request.user,
            source=self.object.project,
            related=self.object.project,
        )

        return response

    def get_success_url(self):
        return self.object.project.get_absolute_url()


class InvoiceAdminView(InvoiceAccessMixin, DelegateableView, DetailView):
    form_views = [ChangeInvoiceStatusView]
    template_name_suffix = "_admin_detail"

    def get_context_data(self, **kwargs):
        invoice = self.get_object()
        project = invoice.project
        deliverables = project.deliverables.all()
        invoice_activities = Activity.actions.filter(
            related_content_type__model="invoice", related_object_id=invoice.id
        ).visible_to(self.request.user)
        return super().get_context_data(
            **kwargs,
            deliverables=deliverables,
            latest_activity=invoice_activities.first(),
            activities=invoice_activities[1:],
        )


class InvoiceApplicantView(InvoiceAccessMixin, DelegateableView, DetailView):
    form_views = []

    def get_context_data(self, **kwargs):
        invoice = self.get_object()
        invoice_activities = Activity.actions.filter(
            related_content_type__model="invoice", related_object_id=invoice.id
        ).visible_to(self.request.user)
        return super().get_context_data(
            **kwargs,
            latest_activity=invoice_activities.first(),
            activities=invoice_activities[1:],
        )


class InvoiceView(ViewDispatcher):
    admin_view = InvoiceAdminView
    finance_view = InvoiceAdminView
    applicant_view = InvoiceApplicantView


class CreateInvoiceView(CreateView):
    model = Invoice
    form_class = CreateInvoiceForm

    def dispatch(self, request, *args, **kwargs):
        self.project = Project.objects.get(pk=kwargs["pk"])
        if not request.user.is_apply_staff and not self.project.user == request.user:
            return redirect(self.project)
        return super().dispatch(request, *args, **kwargs)

    def buttons(self):
        yield ("submit", "primary", _("Save"))

    def get_context_data(self, **kwargs):
        return super().get_context_data(
            project=self.project, buttons=self.buttons(), **kwargs
        )

    def form_valid(self, form):
        form.instance.project = self.project
        form.instance.by = self.request.user

        response = super().form_valid(form)

        if form.cleaned_data["message_for_pm"]:
            invoice_status_change = _("<p>Invoice added.</p>")

            message_for_pm = f'<p>{form.cleaned_data["message_for_pm"]}</p>'

            message = invoice_status_change + message_for_pm

            Activity.objects.create(
                user=self.request.user,
                type=COMMENT,
                source=self.project,
                timestamp=timezone.now(),
                message=message,
                visibility=APPLICANT,
                related_object=self.object,
            )

        messenger(
            MESSAGES.CREATE_INVOICE,
            request=self.request,
            user=self.request.user,
            source=self.project,
            related=self.object,
        )

        if len(self.project.invoices.all()) == 1:
            # remove Project waiting invoices task for applicant on first invoice
            remove_tasks_for_user(
                code=PROJECT_WAITING_INVOICE,
                user=self.project.user,
                related_obj=self.project,
            )

        # add Invoice waiting approval task for Staff group
        add_task_to_user_group(
            code=INVOICE_WAITING_APPROVAL,
            user_group=Group.objects.filter(name=STAFF_GROUP_NAME),
            related_obj=self.object,
        )

        messages.success(
            self.request, _("Invoice added"), extra_tags=PROJECT_ACTION_MESSAGE_TAG
        )

        # Required for django-file-form: delete temporary files for the new files
        # that are uploaded.
        form.delete_temporary_files()
        return response


class EditInvoiceView(InvoiceAccessMixin, UpdateView):
    form_class = EditInvoiceForm

    def dispatch(self, request, *args, **kwargs):
        invoice = self.get_object()
        if not invoice.can_user_edit(request.user):
            return redirect(invoice)
        return super().dispatch(request, *args, **kwargs)

    def buttons(self):
        yield ("submit", "primary", _("Save"))
        if self.object.can_user_delete(self.request.user):
            yield ("delete", "warning", _("Delete"))

    def get_initial(self):
        initial = super().get_initial()
        initial["supporting_documents"] = [
            document.document for document in self.object.supporting_documents.all()
        ]
        return initial

    def get_context_data(self, **kwargs):
        return super().get_context_data(
            project=self.object.project, buttons=self.buttons(), **kwargs
        )

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        form = self.get_form()
        if "delete" in form.data:
            return redirect(
                "apply:projects:invoice-delete",
                pk=self.object.project.id,
                invoice_pk=self.object.id,
            )
        if form.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    def form_valid(self, form):
        old_status = self.object.status
        response = super().form_valid(form)

        if form.cleaned_data:
            if self.object.status in INVOICE_TRANISTION_TO_RESUBMITTED:
                self.object.transition_invoice_to_resubmitted()
                self.object.save()

            if form.cleaned_data["message_for_pm"]:
                invoice_status_change = _(
                    "<p>Invoice status updated to: {status}.</p>"
                ).format(status=self.object.get_status_display())
                message_for_pm = f'<p>{form.cleaned_data["message_for_pm"]}</p>'
                message = invoice_status_change + message_for_pm

                Activity.objects.create(
                    user=self.request.user,
                    type=COMMENT,
                    source=self.object.project,
                    timestamp=timezone.now(),
                    message=message,
                    visibility=APPLICANT,
                    related_object=self.object,
                )

        messenger(
            MESSAGES.UPDATE_INVOICE_STATUS,
            request=self.request,
            user=self.request.user,
            source=self.object.project,
            related=self.object,
        )

        if self.request.user.is_applicant and old_status == CHANGES_REQUESTED_BY_STAFF:
            # remove invoice required changes task for applicant
            remove_tasks_for_user(
                code=INVOICE_REQUIRED_CHANGES,
                user=self.object.project.user,
                related_obj=self.object,
            )

            # add invoice waiting approval task for staff group
            add_task_to_user_group(
                code=INVOICE_WAITING_APPROVAL,
                user_group=Group.objects.filter(name=STAFF_GROUP_NAME),
                related_obj=self.object,
            )

        if (
            self.request.user.is_apply_staff
            and old_status == CHANGES_REQUESTED_BY_FINANCE
        ):
            # remove invoice required changes task for staff group
            remove_tasks_for_user_group(
                code=INVOICE_REQUIRED_CHANGES,
                user_group=Group.objects.filter(name=STAFF_GROUP_NAME),
                related_obj=self.object,
            )
            # add invoice waiting approval task for staff group
            add_task_to_user_group(
                code=INVOICE_WAITING_APPROVAL,
                user_group=Group.objects.filter(name=STAFF_GROUP_NAME),
                related_obj=self.object,
            )

        # Required for django-file-form: delete temporary files for the new files
        # that are uploaded.
        form.delete_temporary_files()
        return response


@method_decorator(login_required, name="dispatch")
class InvoicePrivateMedia(UserPassesTestMixin, PrivateMediaView):
    raise_exception = True

    def dispatch(self, *args, **kwargs):
        invoice_pk = self.kwargs["invoice_pk"]
        project_pk = self.kwargs["pk"]
        self.project = get_object_or_404(Project, pk=project_pk)
        self.invoice = get_object_or_404(self.project.invoices.all(), pk=invoice_pk)

        return super().dispatch(*args, **kwargs)

    def get_media(self, *args, **kwargs):
        file_pk = kwargs.get("file_pk")
        if not file_pk:
            return self.invoice.document

        document = get_object_or_404(self.invoice.supporting_documents, pk=file_pk)
        return document.document

    def test_func(self):
        if self.request.user.is_apply_staff:
            return True

        if self.request.user.is_finance:
            return True

        if self.request.user == self.invoice.project.user:
            return True

        return False


@method_decorator(staff_or_finance_required, name="dispatch")
class InvoiceListView(SingleTableMixin, FilterView):
    filterset_class = InvoiceListFilter
    model = Invoice
    table_class = InvoiceListTable
    template_name = "application_projects/invoice_list.html"
